#!/usr/bin/python3
import hashlib
import logging
import r2pipe
import shutil
import shlex
import sys
import os

logging.basicConfig(
    level=logging.INFO,
    format=' %(levelname)s %(message)s', datefmt='%Y/%m/%d %H:%M:%S'
)
logging.addLevelName(logging.INFO,      '`--')
logging.addLevelName(logging.WARNING,   '<!>')
logging.addLevelName(logging.ERROR,     '<X>')

logger = logging.getLogger(sys.argv[0])


class InvalidArchException(Exception):
    pass


def generate_postinstall(**kwargs):
    with open(os.path.join(os.path.dirname(__file__), "postinstall.template"), "r") as fin:
        content = fin.read()
    for k, v in kwargs.items():
        content = content.replace("{{" + k + "}}", str(v))
    return content


def patch_pdfm_srv(pdfm_srv_path, arch = "arm"):
    p1 = p2 = -1
    if arch == "arm":
        asm = "mov x0, 0x1; ret;"
    elif arch == "x86":
        asm = "push 0x1; pop rax; ret;"
    else:
        raise InvalidArchException("arch should be either 'arm' or 'x86'")
    r2 = r2pipe.open(pdfm_srv_path, flags=["-2", "-a", arch])
    # Verify binary architecture
    for b in r2.cmdj(f"!rabin2 -j -A {shlex.quote(pdfm_srv_path)}")['bins']:
        if b['arch'] == arch:
            break
    else:
        raise InvalidArchException(f"arch '{arch}' not found in this binary")
    # Run binary analysis
    r2.cmd("aaa")
    # Calculate virtual address -> binary offset
    file_map    = r2.cmdj("om.j")
    offset      = file_map["delta"] - file_map["from"]
    # Recover class info from RTTI, then find the patch points
    r2.cmd("avrr")
    p1res1      = r2.cmdj("acllj DspLic::SignCheckerImpl")["methods"]
    p1res2      = r2.cmdj("acllj Signature::SignCheckerImpl")["methods"]
    p2res1      = r2.cmdj("axtj str._usr_bin_codesign")
    if p1res1:
        p1 = p1res1[-1]["addr"]
    elif p1res2:
        p1 = p1res2[-1]["addr"]
    if p2res1:
        p2 = p2res1[-1]["fcn_addr"]
    # Write patch code
    if p1 >= 0:
        r2.cmd(f"oo+")
        r2.cmd(f"s {p1}")
        r2.cmd(f'"wa {asm}"')
    if p2 >= 0:
        r2.cmd(f"oo+")
        r2.cmd(f"s {p2}")
        r2.cmd(f'"wa {asm}"')
    r2.quit()
    return p1 + offset, p2 + offset


def main():
    if len(sys.argv) < 3:
        print(f"Usage: python3 {sys.argv[0]} VERSION /path/to/prl_disp_service /path/to/output_dir\n")
        return
    full_version        = sys.argv[1]
    pdfm_srv_in_path    = sys.argv[2]
    output_dir          = sys.argv[3]
    pdfm_srv_out_path   = os.path.join(output_dir, "prl_disp_service.patched")
    shutil.copy(pdfm_srv_in_path, pdfm_srv_out_path)

    a01 = a02 = i01 = i02 = -1
    for arch in "arm", "x86":
        try:
            logger.info(f"Patching {arch} code...")
            x01, x02 = patch_pdfm_srv(pdfm_srv_out_path, arch)
            if x01 < 0 or x02 < 0:
                logger.warning(f"One or more patch points not found in {arch} {(x01, x02)}")
            if arch == "arm":
                a01, a02 = x01, x02
            elif arch == "x86":
                i01, i02 = x01, x02
        except InvalidArchException as ex:
            logger.warning(str(ex))

    logger.info(f"Generating postinstall script...")
    with open(pdfm_srv_in_path, "rb") as fin:
        md5_before = hashlib.md5(fin.read()).hexdigest()
    with open(pdfm_srv_out_path, "rb") as fin:
        md5_after = hashlib.md5(fin.read()).hexdigest()
    if md5_before == md5_after:
        raise RuntimeError("Nothing is patched!")
    os.makedirs(os.path.join(output_dir, "scripts"), exist_ok=True)
    with open(os.path.join(output_dir, "scripts", "postinstall"), "w") as fout:
        fout.write(generate_postinstall(
            full_version    = full_version,
            hash_before     = md5_before,
            hash_after      = md5_after,
            i_01            = i01,
            i_02            = i02,
            a_01            = a01,
            a_02            = a02
        ))
    os.chmod(os.path.join(output_dir, "scripts", "postinstall"), mode=0o755)
    logger.info(f"Patch info: {i01}/{i02}/{a01}/{a02}")


if __name__ == '__main__':
    main()
